/*---------------------------------------------------------------
 *	MMLの楽譜を2台で合奏(μT-Kernel 3.0用)
 *
 *	Copyright (C) 2022-2024 by T3 WG of TRON Forum
 *---------------------------------------------------------------*/
#include <tk/tkernel.h>
#include <tm/tmonitor.h>

//-------- UARTEのレジスタ定義 ----------------------------------
#define	UARTE(p, r)	(UARTE##p##_BASE + UARTE_##r)

#define	UARTE0_BASE	0x40002000	// コンソール用UART0と競合するので使用しない
#define	UARTE1_BASE	0x40028000	// こちらを使用する

#define	UARTE_TASKS_STARTRX		(0x000)
#define	UARTE_TASKS_STOPRX		(0x004)
#define	UARTE_TASKS_STARTTX		(0x008)
#define	UARTE_TASKS_STOPTX		(0x00C)
#define	UARTE_TASKS_FLUSHRX		(0x02C)

#define	UARTE_EVENTS_CTS		(0x100)
#define	UARTE_EVENTS_NCTS		(0x104)
#define	UARTE_EVENTS_RXDRDY		(0x108)
#define	UARTE_EVENTS_ENDRX		(0x110)
#define	UARTE_EVENTS_TXDRDY		(0x11C)
#define	UARTE_EVENTS_ENDTX		(0x120)
#define	UARTE_EVENTS_ERROR		(0x124)
#define	UARTE_EVENTS_RXTO		(0x144)

#define	UARTE_EVENTS_RXSTARTED	(0x14C)
#define	UARTE_EVENTS_TXSTARTED	(0x150)
#define	UARTE_EVENTS_TXSTOPPED	(0x158)

#define	UARTE_SHORTS			(0x200)
#define	UARTE_INTEN				(0x300)
#define	UARTE_INTENSET			(0x304)
#define	UARTE_INTENCLR			(0x308)
#define	UARTE_ERRORSRC			(0x480)
#define	UARTE_ENABLE			(0x500)
#define	UARTE_PSEL_RTS			(0x508)
#define	UARTE_PSEL_TXD			(0x50C)
#define	UARTE_PSEL_CTS			(0x510)
#define	UARTE_PSEL_RXD			(0x514)
#define	UARTE_BAUDRATE			(0x524)

#define	UARTE_RXD_PTR			(0x534)
#define	UARTE_RXD_MAXCNT		(0x538)
#define	UARTE_RXD_AMOUNT		(0x53C)
#define	UARTE_TXD_PTR			(0x544)
#define	UARTE_TXD_MAXCNT		(0x548)
#define	UARTE_TXD_AMOUNT		(0x54C)

#define	UARTE_CONFIG			(0x56C)

//-------- PWMのレジスタ定義 ------------------------------------
#define PWM_BASE(un)	((un) == 0 ? 0x4001c000 : (un) == 1 ? 0x40021000 : (un) == 2 ? 0x40022000 : 0x4002d000)
#define PWM(un, r)				(PWM_BASE(un) + PWM_##r)
#define PWM_TASKS_STOP			0x004
#define PWM_TASKS_SEQSTART(n)	(0x008 + (n) * 0x04)
#define PWM_LOOPSDONE			0x11C
#define PWM_ENABLE				0x500
#define PWM_MODE				0x504
#define PWM_COUNTERTOP			0x508
#define PWM_PRESCALER			0x50c
#define PWM_DECODER				0x510
#define PWM_LOOP				0x514
#define PWM_SEQ_PTR(n)			(0x520 + (n) * 0x20)
#define PWM_SEQ_CNT(n)			(0x524 + (n) * 0x20)
#define PWM_SEQ_REFRESH(n)		(0x528 + (n) * 0x20)
#define PWM_SEQ_ENDDELAY(n)		(0x52c + (n) * 0x20)
#define PWM_PSEL_OUT(n)			(0x560 + (n) * 0x04)

//-------- ミリ秒単位の現在時刻の下位32ビットを返す -------------
LOCAL W cur_time(void)
{
	SYSTIM	tim;
	tk_get_tim(&tim);			// 現在時刻を取得
	return(tim.lo);
}

//-------- ボードIDの取得 ---------------------------------------
#define	NRF5_FICR_BASE	0x10000000
#define	DEVICEID_0		0x060		// Device identifier 0
#define	DEVICEID_1		0x064		// Device identifier 1

// メインボード用micro:bitの実際のデバイスID
const W main_devid0 = 0x01234567;	// (★要設定★)
const W main_devid1 = 0x89abcdef;	// (★要設定★)

W devid0, devid1;					// 実行中のmicro:bitのデバイスID

// 自分自身がメインボードの場合にTRUEを返す関数(※B)
LOCAL BOOL is_main_board(){
	return(devid0 == main_devid0 && devid1 == main_devid1);
}

// ボードのデバイスIDを取得してコンソールに表示
LOCAL void get_device_id(void)
{
	devid0 = in_w(NRF5_FICR_BASE + DEVICEID_0);	// DEVICE[0]の読み出し(※A)
	devid1 = in_w(NRF5_FICR_BASE + DEVICEID_1);	// DEVICE[1]の読み出し(※A)

	tm_printf("This devid_0_1: %08x_%08x\n", devid0, devid1);
	tm_printf("Main devid_0_1: %08x_%08x\n", main_devid0, main_devid1);
	if(is_main_board())
		tm_printf("This is MAIN board, P0TX_P1RX\n\n");
	else
		tm_printf("This is  SUB board, P1TX_P0RX\n\n");
}	

//-------- LED制御のためのGPIOの初期設定 ------------------------
LOCAL void led_init(void)
{
	// COL1..COL5とROW1..ROW5に接続されたGPIOピンを出力に設定
	out_w(GPIO(P0, PIN_CNF(28)), 1);	// GPIO P0.28(COL1に接続)
	out_w(GPIO(P0, PIN_CNF(11)), 1);	// GPIO P0.11(COL2に接続)
	out_w(GPIO(P0, PIN_CNF(31)), 1);	// GPIO P0.31(COL3に接続)
	out_w(GPIO(P1, PIN_CNF(05)), 1);	// GPIO P1.05(COL4に接続)
	out_w(GPIO(P0, PIN_CNF(30)), 1);	// GPIO P0.30(COL5に接続)
	out_w(GPIO(P0, PIN_CNF(21)), 1);	// GPIO P0.21(ROW1に接続)
	out_w(GPIO(P0, PIN_CNF(22)), 1);	// GPIO P0.22(ROW2に接続)
	out_w(GPIO(P0, PIN_CNF(15)), 1);	// GPIO P0.15(ROW3に接続)
	out_w(GPIO(P0, PIN_CNF(24)), 1);	// GPIO P0.24(ROW4に接続)
	out_w(GPIO(P0, PIN_CNF(19)), 1);	// GPIO P0.19(ROW5に接続)
}

//-------- GPIOの特定のピンpinの出力をval(0または1)に設定 -------
//	pinで指定された以外のピンの状態は変化しない
LOCAL void out_gpio_pin(UW port, UW pin, UW val)
{
	INT port_addr;

	if(port == 1){							// P1の場合
		if(val)
			port_addr = GPIO(P1, OUTSET);	// P1でvalが1の場合
		else
			port_addr = GPIO(P1, OUTCLR);	// P1でvalが0の場合

	} else {								// P0の場合
		if(val)
			port_addr = GPIO(P0, OUTSET);	// P0でvalが1の場合
		else
			port_addr = GPIO(P0, OUTCLR);	// P0でvalが0の場合
	}
	out_w(port_addr, (1 << pin));			// 指定のピンに1を出力
}

//-------- ボタンスイッチに対応するGPIO P0のピンを入力に設定 ----
LOCAL void btn_init(void)
{
	out_w(GPIO(P0, PIN_CNF(14)), 0);	// ボタンスイッチAはGPIO P0.14に接続
	out_w(GPIO(P0, PIN_CNF(23)), 0);	// ボタンスイッチBはGPIO P0.23に接続
}

// UARTEの送信データと受信データを入れるバッファメモリ
#define	MAXBUF	2
static _UB	txbuf[MAXBUF];
static _UB	rxbuf[MAXBUF];

LOCAL BOOL	mml_playing = FALSE;	// MML演奏中を示すフラグ
LOCAL ID	flgid;					// 割込み通知用イベントフラグのID
LOCAL ID	semid;					// 送信バッファ排他制御用セマフォのID

//-------- 割込み許可を含めたUARTE1の初期化とピン設定 -----------
//			最後の割込み許可以外は連載第13回のinit_uarte1と同じ
LOCAL void init_uarte1(INT pin_txd, INT pin_rxd)
{
	out_w(UARTE(1, EVENTS_ENDRX), 0);
	out_w(UARTE(1, EVENTS_ENDTX), 0);
	out_w(UARTE(1, EVENTS_ERROR), 0);
	out_w(UARTE(1, ERRORSRC), 0);

	out_w(UARTE(1, BAUDRATE), 0x00275000);		// 9600 baud(※B)
	// out_w(UARTE(1, BAUDRATE), 0x01D7E000);	// 115200 baud(※B)

	// Hardware flow control無し、パリティ無し、One stop bit
	out_w(UARTE(1, CONFIG), 0);

	out_w(UARTE(1, PSEL_TXD), pin_txd);	// TXDのGPIOピン設定(※C)
	out_w(UARTE(1, PSEL_RXD), pin_rxd);	// RXDのGPIOピン設定(※C)
	out_w(UARTE(1, ENABLE), 8);			// Enable UARTE(※D)

	// EVENTS_ENDTXとEVENTS_ENDRXの割込み許可の設定
	out_w(UARTE(1, INTEN), 0x00000110);
}

//-------- UARTE1からtxdatの1バイトを送信開始(※C) --------------
//			ボタン状態変化により呼ばれる
LOCAL void uarte1_start_tx1(UB txdat){

	tk_wai_sem(semid, 1, TMO_FEVR);			// 送信バッファの排他制御(※D)

	txbuf[0] = txdat;						// 送信データを送信バッファに格納

	out_w(UARTE(1, TXD_PTR), (UW) txbuf);	// 送信バッファ先頭アドレス
	out_w(UARTE(1, TXD_MAXCNT), 1);			// 送信するバイト数

	out_w(UARTE(1, EVENTS_ENDTX), 0);		// 送信終了フラグのクリア
	out_w(UARTE(1, TASKS_STARTTX), 1);		// 送信開始
}

//-------- UARTE1で1バイトの受信開始(※E) -----------------------
//			起動後および受信完了の割込みハンドラから呼ばれる
LOCAL void uarte1_start_rx1(){
	out_w(UARTE(1, RXD_PTR), (UW) rxbuf);	// 受信バッファ先頭アドレス
	out_w(UARTE(1, RXD_MAXCNT), 1);			// 受信するバイト数

	out_w(UARTE(1, EVENTS_ENDRX), 0);		// 受信終了フラグのクリア
	out_w(UARTE(1, TASKS_STARTRX), 1);		// 受信開始
}

//-------- UARTE1の送信完了と受信完了の割込みハンドラ -----------
//			送信完了と受信完了が同時に起こる可能性もあるので、
//			8ビットの送信データと受信データを1つの16ビットデータに
//			まとめてからset_flgで送る
LOCAL void uarte1_inthdr(UINT intno)
{
	UW		txrxdat = 0;	// 受信データと送信データを入れてset_flgで送る変数
	_UB*	txrxadr;		// 送信バッファまたは受信バッファの先頭アドレス

	if(in_w(UARTE(1, EVENTS_ENDRX))){		// 受信完了の割込みの場合(※F)
		txrxadr = (_UB *) in_w(UARTE(1, RXD_PTR));
		txrxdat |= (UW) txrxadr[0];			// 下位8ビットに受信データを格納
		uarte1_start_rx1();					// 次の受信の準備(※G)
	}

	if(in_w(UARTE(1, EVENTS_ENDTX))){		// 送信完了の割込みの場合(※H)
		txrxadr = (_UB *) in_w(UARTE(1, TXD_PTR));

		// txrxdatの上位8ビットに送信データを格納(※J)
		txrxdat |= ((UW) txrxadr[0] << 8);

		out_w(UARTE(1, EVENTS_ENDTX), 0);	// 送信終了フラグのクリア(※K)
		tk_sig_sem(semid, 1);				// 送信バッファの排他制御を解除(※L)
	}

	tk_set_flg(flgid, txrxdat);				// 受信データと送信データの通知(※M)

	mml_playing = FALSE;					// MML演奏中だった場合は中止
}

// 以前のボタンの状態を保持する変数
LOCAL BOOL prev_a = FALSE;
LOCAL BOOL prev_b = FALSE;

//---------------------------------------------------------------
// ボタンスイッチの状態をチェック(※P)
LOCAL void check2_btn(BOOL new_btn, BOOL *prev_btn, UB btnchr){
	UB actchr;

	if(new_btn ==  *prev_btn) return;	// 状態変化が無ければ何もせずに戻る

	*prev_btn = new_btn;	

	if(new_btn){						// ボタンが押された場合
		uarte1_start_tx1(btnchr);		// btnchrのデータをUARTE1で送信
	}
}

// ボタンスイッチの入力判定を行う送信側タスク(※N)
LOCAL void btn_tx_task(INT stacd, void *exinf){
	UW gpio_p0in;
	BOOL btn_a, btn_b;

	for (;;) {				// 約100msごとに永久に繰り返し

		// GPIO P0のINレジスタを読んでgpio_p0inに設定
		gpio_p0in = in_w(GPIO(P0, IN));

		btn_a = ((gpio_p0in & (1 << 14)) == 0); // P0.14が0の場合にTRUE
		btn_b = ((gpio_p0in & (1 << 23)) == 0); // P0.23が0の場合にTRUE

		check2_btn(btn_a, &prev_a, 'A');	// ボタンAの変化のチェックと送信
		check2_btn(btn_b, &prev_b, 'B');	// ボタンBの変化のチェックと送信

		tk_dly_tsk(100);					// 100msのディレイ
	}
}

// -------------------------------------------------------
// LED用PWMのカウンタ上限値の定義(周期時間は1マイクロ秒)
#define	TOP				1000

const INT led_pwm = 0;				 	// unit=0のPWMを使ってLEDを点灯

// -------------------------------------------------------
// LED用PWMのカウンタ比較値を格納する3次元配列
//	comp_table[パターンptn][ステップstep][チャンネルch]
//	&comp_table[ptn][0][0] のアドレスをSEQ_PTR(0,1)に設定
/*
  --------------------------------------------------------------
								ch=0	ch=1	ch=2	ch=3
								 ↓		 ↓		 ↓		 ↓
  --------------------------------------------------------------
	ptn=0	step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
	ptn=1	step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
	ptn=2
								...
  --------------------------------------------------------------
								...
  --------------------------------------------------------------
	ptn=(MAX_PTN-1)
			step=0				COMP0	COMP1	COMP2	COMP3
			step=1				COMP0	COMP1	COMP2	COMP3
			...					...		...		...		...
			step=(MAX_STEP-1)	COMP0	COMP1	COMP2	COMP3
  --------------------------------------------------------------
*/
#define	MAX_PTN			20
#define	PTN_01_OFFSET	10			// SEQ0用ptnとSEQ1用ptnのインデックスの差
#define	PTN_DOREMI_CNT	8			// 音階に対応して使用するパターン数
#define	PTN_SEQ0_MIN	0									// SEQ0用ptnの最小値
#define	PTN_SEQ0_MAX	(PTN_SEQ0_MIN + PTN_DOREMI_CNT - 1)	// SEQ0用ptnの最大値
#define	PTN_SEQ1_MIN	(PTN_SEQ0_MIN + PTN_01_OFFSET)		// SEQ1用ptnの最小値
#define	PTN_SEQ1_MAX	(PTN_SEQ1_MIN + PTN_DOREMI_CNT - 1)	// SEQ1用ptnの最大値
#define	PTN_SEQ0_OFF	(PTN_SEQ0_MAX + 1)		// SEQ0用の点灯OFFパターン
#define	PTN_SEQ1_OFF	(PTN_SEQ1_MAX + 1)		// SEQ1用の点灯OFFパターン

#define	MAX_STEP	16
#define	MAX_CH		4
#define	SEQ0_PERCENT	10		// (SEQ0+SEQ1)に対するSEQ0の時間の割合(%)

// カウンタ比較値を入れる配列
static _UH comp_table[MAX_PTN][MAX_STEP][MAX_CH];

// 千分率のdutyからカウンタ比較値への変換
LOCAL UH duty2comp(INT duty){
	return(TOP - (duty * TOP / 1000));
}

// 指定のパターン、チャンネル、最大デューティ比(max_duty)と減衰率(decay)から、
//	各ステップの比較値をcomp_tableに設定
//	max_dutyとdecayは千分率で指定
//
//	(dir > 0 の場合)
//	comp_table[][step=0][] に max_duty を設定
//	comp_table[][step=1][] に max_duty * (decay / 1000) を設定
//	comp_table[][step=2][] に max_duty * (decay / 1000)^2 を設定
//	comp_table[][step=3][] に max_duty * (decay / 1000)^3 を設定
//			...
//
//	(dir < 0 の場合)
//	comp_table[][step=MAX_STEP-1][] に max_duty を設定
//	comp_table[][step=MAX_STEP-2][] に max_duty * (decay / 1000) を設定
//	comp_table[][step=MAX_STEP-3][] に max_duty * (decay / 1000)^2 を設定
//	comp_table[][step=MAX_STEP-4][] に max_duty * (decay / 1000)^3 を設定
//			...
//
LOCAL void init_step_duty(INT max_duty, INT ptn, INT step_decay,
							INT ch, INT dir){
	for(INT idx = 0; idx < MAX_STEP; idx++){
		comp_table[ptn][dir > 0 ? idx : (MAX_STEP-1-idx)][ch]
			= duty2comp(max_duty);
		max_duty = max_duty * step_decay / 1000;
	}
}

// 距離(差異)に応じて千分率による減衰率を返す
LOCAL INT ch_decay(INT dist){

	// 負数の場合もあるので2乗してから比較
	switch (dist * dist){
		case (0*0):	return(1000);
		case (1*1):	return(300);
		case (2*2):	return(100);
		case (3*3):	return(50);
		case (4*4):	return(10);
		case (5*5):	return(5);
		case (6*6):	return(1);
		case (7*7):	return(0);
		default:
			tm_printf("ch_decay: illegal dist=%d\n", dist);
			return(0);
	}
}

// ドレミの音階へのインデックスdoremiに応じて指定のパターンを生成
//	チャンネル番号(0..3)の2倍がちょうど音階へのインデックス(=0..7)に
//	なるチャンネルでは、max_dutyをそのまま使う
//	その前後のチャンネルでは、両者の距離(差異)に応じてch_decayによりdutyが減衰
LOCAL void init_ptn_duty(INT max_duty, INT ptn,
							INT step_decay, INT doremi, INT dir){
	for(INT idx = 0; idx < MAX_CH; idx++){
		init_step_duty(ch_decay(idx * 2 - doremi) * max_duty / 1000,
			ptn, step_decay, idx, dir);
	}
}

// すべてのパターン番号に対するcomp_tableを設定
LOCAL void init_pwm_duty(){
	INT idx_ptn;

	// ドレミの音階に応じたSEQ0用のパターンの生成
	for(idx_ptn = PTN_SEQ0_MIN; idx_ptn <= PTN_SEQ0_MAX; idx_ptn++){
		init_ptn_duty(1000, idx_ptn, 400, idx_ptn, (-1));
	}

	// ドレミの音階に応じたSEQ1用のパターンの生成
	for(idx_ptn = PTN_SEQ1_MIN; idx_ptn <= PTN_SEQ1_MAX; idx_ptn++){
		init_ptn_duty(1000, idx_ptn, 700, (idx_ptn - PTN_01_OFFSET), 1);
	}

	// すべて消灯するパターン(PTN_SEQ0_OFF,PTN_SEQ1_OFF)の生成
	init_ptn_duty(0, PTN_SEQ0_OFF, 0, 0, 1);
	init_ptn_duty(0, PTN_SEQ1_OFF, 0, 0, 1);
}

// 比較値のテーブルを設定してからPWMをスタート
LOCAL void pwm_duty_start(INT unit,
	_UH *seq0_table, _UH *seq1_table, INT seq0_refresh, INT seq1_refresh){

  // tm_printf("pwm_duty_start: seq0tab=%08x, seq1tab=%08x, seq0ref=%d, seq1ref=%d\n", seq0_table, seq1_table, seq0_refresh, seq1_refresh);

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);
	out_w(PWM(unit, PRESCALER), 4);		// DIV_16でクロック1MHzを指定
	out_w(PWM(unit, COUNTERTOP), TOP);	// カウンタ上限値の設定
	out_w(PWM(unit, DECODER), 2);		// LOAD=individual(single)を指定(※D)

	// シーケンス0の比較値の配列の先頭アドレスseq0_tableを設定
	out_w(PWM(unit, SEQ_PTR(0)), (UW)seq0_table);
	out_w(PWM(unit, SEQ_CNT(0)), MAX_STEP * MAX_CH);
	out_w(PWM(unit, SEQ_REFRESH(0)), seq0_refresh);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	// シーケンス1の比較値の配列の先頭アドレスseq1_tableを設定
	out_w(PWM(unit, SEQ_PTR(1)), (UW)seq1_table);
	out_w(PWM(unit, SEQ_CNT(1)), MAX_STEP * MAX_CH);
	out_w(PWM(unit, SEQ_REFRESH(1)), seq1_refresh);
	out_w(PWM(unit, SEQ_ENDDELAY(1)), 0);

	// 繰り返し1回を指定してPWMのシーケンス0の動作を開始
	out_w(PWM(unit, LOOP), 1);
	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// -------------------------------------------------------
//	PWMを使ったスピーカーの制御
// -------------------------------------------------------
#define SPEAKER_GPIO_PORT	0	// SPEAKERに接続されたGPIOのポート番号
#define SPEAKER_GPIO_PIN	0	// SPEAKERに接続されたGPIOのピン番号

// 比較値COMP0を格納する16ビット幅のメモリ領域
static _UH half_cycle_us;

// unitで指定されたPWMに周波数freq(Hz)の発振を設定
LOCAL void pwm_speaker_start(INT unit, INT freq)
{
	// 1周期の波形に対するマイクロ秒単位の時間を計算
	// (＝1MHzのクロックに対するカウント値)
	INT cycle_us = (1000 * 1000) / freq;

	// cycle_usの半分の時間(＝1MHzのクロックに対するカウント値の半分)を計算
	half_cycle_us = cycle_us / 2;

	out_w(PWM(unit, ENABLE), (1 << 0));
	out_w(PWM(unit, MODE), 0);

	out_w(PWM(unit, PRESCALER), 4);	// クロック1MHzで利用(Divide by 16)

	out_w(PWM(unit, COUNTERTOP), cycle_us);
	out_w(PWM(unit, LOOP), 0);
	out_w(PWM(unit, DECODER), 0);	// LOAD=commonを指定

	out_w(PWM(unit, SEQ_PTR(0)), (UW)&half_cycle_us);
	out_w(PWM(unit, SEQ_CNT(0)), 1);
	out_w(PWM(unit, SEQ_REFRESH(0)), 0);
	out_w(PWM(unit, SEQ_ENDDELAY(0)), 0);

	out_w(PWM(unit, TASKS_SEQSTART(0)), 1);
}

// unitで指定されたPWMの発振を停止
LOCAL void pwm_stop(INT unit)
{
	out_w(PWM(unit, TASKS_STOP), 1);
}

//---------------------------------------------------------------
// MMLによる楽譜データの処理
//	MML(Music Macro Language)の仕様にほぼ準拠するが未実装の機能もある
//
//	書式: {<cmd>[<sharp>][<num>][<ldot>]}*
//
//	<cmd>:		1文字のみ、英文字は大文字でもよい
//		t		テンポの指定
//		v		音量の指定(無視される)
//		@		音色の指定(無視される)
//		q		音符を発音する割合の指定(無視される)
//		l		デフォルトの音符の長さの指定
//		o		オクターブの指定
//		>		オクターブを1つ上げる指定
//		<		オクターブを1つ下げる指定
//		c,d,e,f,g,a,b	ドレミファソラシの音符
//		r		休符
//
//	<sharp>:	1文字のみ
//		#,+		半音上げる
//		-		半音下げる
//
//	<num>:		[0-9]*、複数文字可、10進数で表現された数字
//				音符の場合は "<num>分音符" を表す
//
//	<ldot>:		[/._]*、複数文字可
//		/		音符の長さを半分にする
//		.		音符の長さを1.5倍にする
//		_		音符の長さを2倍にする
//

//---------------------------------------------------------------
const INT speaker_pwm = 1;		// スピーカー用PWMとしてunit=1を使用

#define	SHARP_FREQ	10595		// 半音上下する場合の10000との周波数比

INT	tempo, octave, lnum, volume;
INT	prev_time;					// 音符の再生を開始した時刻

// ドレミの音階の周波数(Hz): C5,D5,E5,F5,G5,A5,B5)
const INT pwm_freqs[] = { 523, 587, 659, 698, 783, 880, 987 };

LOCAL INT mmlcnt;				// 再生済の音符数、進捗状況

//---------------------------------------------------------------
LOCAL BOOL is_num(B ch){				// chが数字の場合にTRUEを返す
	return(ch >= '0' && ch <= '9');
}
	
LOCAL BOOL is_chars(B chr, B* chars){	// chrがcharsに含まれる場合にTRUEを返す
	while(*chars != '\0'){
		if(chr == *chars++)
		return(TRUE);
	}
	return(FALSE);
}
	
// 1つの音符や休符の再生
LOCAL void play_one_mml(B cmd, B sharp, INT olen_num, INT ldot){
	INT index_freqs = cmd - 'c';
	INT play_freq, led_ptn, play_time;
	INT	seq0_time, seq1_time, seq0_refresh, seq1_refresh;
	_UH	*seq0_table, *seq1_table;

	if(index_freqs < 0)	index_freqs += 7;	// cmdが 'a','b' の場合

	if(cmd == 'r'){							// 休符の場合

	    play_freq = 0;
		seq0_table = &comp_table[PTN_SEQ0_OFF][0][0]; 	// LEDを消灯
		seq1_table = &comp_table[PTN_SEQ1_OFF][0][0];

	} else {								// 休符以外の音符の場合

		play_freq = pwm_freqs[index_freqs];	// cmdの音符に対応する周波数
		if(sharp == '+' || sharp == '#')				// 半音上げる
			play_freq = play_freq * SHARP_FREQ / 10000;
		else if(sharp == '-')							// 半音下げる
			play_freq = play_freq * 10000 / SHARP_FREQ;
		if(octave > 5)									// オクターブ上げる
			play_freq = play_freq << (octave - 5);
		else if(octave < 5)								// オクターブ下げる
			play_freq = play_freq >> (5 - octave);

		// LED用PWMのパラメータとしてptnに応じた点灯パターンを設定
		led_ptn = index_freqs;
		seq0_table = &comp_table[PTN_SEQ0_MIN + led_ptn][0][0];
		seq1_table = &comp_table[PTN_SEQ1_MIN + led_ptn][0][0];

		// スピーカー用PWMを制御して音符を再生
		pwm_speaker_start(speaker_pwm, play_freq);
	}

	if (olen_num == 0)	olen_num = lnum;	// 音符長の指定がない場合はlnum

	// tempoは1分間(=60000ms)の4分音符の数なので、
	// ms単位での音符の再生時間play_timeは以下のように計算できる
	play_time = 60000 * 4 * ldot / tempo / olen_num / 256;

	tm_printf("%5d: play_time=%d, play_freq=%d\n",
		mmlcnt, play_time, play_freq);
 
	seq0_time = play_time * SEQ0_PERCENT / 100;	// ミリ秒単位のSEQ0時間
	seq1_time = play_time - seq0_time;			// ミリ秒単位のSEQ1時間

	// SEQ0,SEQ1のミリ秒単位の時間からREFRESHレジスタの設定値を計算
	//	1ステップのサイクル数 = ミリ秒単位の1ステップの所要時間
	//							= (seq0_time / MAX_STEP)
	seq0_refresh =	seq0_time / MAX_STEP - 1;
	if(seq0_refresh < 0) seq0_refresh = 0;		// (-1)になった場合は0に戻す
	seq1_refresh =	seq1_time / MAX_STEP - 1;
	if(seq1_refresh < 0) seq1_refresh = 0;		// (-1)になった場合は0に戻す

	// LED用PWMを制御してLEDを点灯
	pwm_duty_start(led_pwm, seq0_table, seq1_table, seq0_refresh, seq1_refresh);

	prev_time += play_time;				// 音符の再生を終了すべき時刻の計算
	mmlcnt++;

	// 終了すべき時刻までディレイ
	tk_dly_tsk(prev_time - cur_time() - 10);

	pwm_stop(speaker_pwm);				// 音符の再生を終了
	tk_dly_tsk(10);						// 各音符の後のディレイ
}

// MMLの各コマンドの実行
LOCAL void exec_one_mml(B cmd, B sharp, INT num, INT ldot){

	tm_printf("%5d: cmd=%c, sharp=%c, num=%d, dot= %d\n",
		mmlcnt, cmd, sharp, num, ldot);

	if(cmd >= 'A' && cmd <= 'Z')		// cmdが英文字の場合に小文字に変換
		cmd += ('a' - 'A');

	if((cmd >= 'a' && cmd <= 'g') || cmd == 'r'){	// 1つの音符や休符の再生
		play_one_mml(cmd, sharp, num, ldot);
		return;
	}
	
	// 音符や休符以外の<cmd>の実行
	switch(cmd){
		case 't':		// テンポの設定
			tempo = num;	break;
		case 'o':		// オクターブの設定
			octave = num;	break;
		case 'l':		// 指定がない場合の音符の長さ
			lnum = num;	break;
		case 'v':		// 音量の指定
			volume = num;	break;
		case '>':		// 1オクターブ上げる
			octave++;	break;
		case '<':		// 1オクターブ下げる
			octave--;	break;
		default:		// 定義されていない文字は無視、何もしない
		  	break;
	}
}

// MMLの文字列から1つの <cmd> を切り取ってexec_one_mmlを実行
//	その次の <cmd> を指すポインタを返す
LOCAL B* read_mml(B *mml){
	INT	num = 0, ldot = 256;	// exec_one_mml実行時のデフォルト値
	B	cmd, sharp = 'x';

	if(mml[0] == 'M' && mml[1] == 'M' && mml[2] == 'L' && mml[3] == '@')
	    mml += 4;			// 最初が "MML@" だった場合に読み飛ばす

	cmd = *mml++;

	if(is_chars(*mml, "#+-"))	// <sharp>がある場合にsharpを設定
		sharp = *mml++;

	while(is_num(*mml))			// <num>がある場合に設定(無ければ0のまま)
    	num = num * 10 + (*mml++ - '0');

	while(is_chars(*mml, "/._")){	// <ldot>がある場合に設定(無ければ256のまま)
		if(*mml == '/')	ldot = ldot / 2;
		if(*mml == '.')	ldot = ldot / 2 * 3;
		if(*mml == '_')	ldot = ldot * 2;
		mml++;
	}

	exec_one_mml(cmd, sharp, num, ldot);	// コマンド実行
	return(mml);
}

// MML全体の実行
LOCAL void play_all_mml(B *mml){

	mml_playing = TRUE;		// MML演奏中
	tempo = 120;			// テンポの指定、初期値は120
	octave = 4;				// オクターブ指定、初期値は4
	lnum = 4;				// 指定がない場合の音符長、初期値は4分音符
	volume = 8;				// 音量の指定(使用しない)

	prev_time = cur_time();	// 開始時刻の設定
	mmlcnt = 1;

	for(;;){
	    mml = read_mml(mml);
	    if(! mml_playing)	return;		// MML演奏中止
	    if(*mml == '\0')	return;		// MMLの楽譜の最後
	}
}

//---------------------------------------------------------------
//	MMLの楽譜データ
//		mm_XXX がメインボード用(主旋律)
//		ms_XXX がサブボード用(伴奏)

// 起動時のチャイム
B mm_start[] = "t720>ceg>c2.";
B ms_start[] = "t720>>c<gec2.";

// きらきら星
B mm_kira[] = "ccggaag_ ffeeddc_ ggffeed_ ggffeed_ ccggaag_ ffeeddc_";
B ms_kira[] = "ccccffc_ ddcc<bb>c_ ccddcc<b_> ccddcc<b_> l8c<a>c<a>ececfdfdedc<b> d<b>d<b>c<a>c<a><baba>c__";

// かえるの合唱(輪唱)
B mm_kaeru[] = "     cdefedc.r8 efgagfe.r8 crcrcrcr l8ccddeeffl4edc.r8";
B ms_kaeru[] = "r1r1 cdefedc.r8 efgagfe.r8 crcrcrcr l8ccddeeffl4edc.r8";

// メインボード用のMMLの楽譜を入れた文字列の配列(※T)
B *mml_main[] = { mm_kira, mm_kaeru };

// サブボード用のMMLの楽譜を入れた文字列の配列(※U)
B *mml_sub[]  = { ms_kira, ms_kaeru };

// 演奏中の楽譜のインデックス(曲番号)
LOCAL INT mml_idx = 0;

// mml_idxで指定された楽譜の1曲を演奏(※V)
LOCAL void play_mml_idx(void){

	if(is_main_board()){
		tm_printf("play_mml_idx: mml_idx=%d (MAIN)\n", mml_idx);
		play_all_mml(mml_main[mml_idx]);
	} else {
		tm_printf("play_mml_idx: mml_idx=%d (SUB)\n", mml_idx);
		play_all_mml(mml_sub[mml_idx]);
	}
	tm_printf("play_mml_idx: mml_idx=%d -- END\n", mml_idx);
}

// 表示できない文字を '.' に変換する関数
LOCAL UINT disp_ch(UINT ch){
	ch &= 0xff;
	if(ch < ' ' || ch > '~')	return('.');
	else						return(ch);
}

// イベントフラグによりPWMのMML演奏を開始する受信側タスク
LOCAL void flg_rx_task(INT stacd, void *exinf){
	UINT flgptn;
	UB	txchr, rxchr, cmdchr;

	if(is_main_board())
		play_all_mml(mm_start);
	else
		play_all_mml(ms_start);

	for (;;) {				// 永久に繰り返し

		// 割込み通知用イベントフラグの下位2バイトをOR待ち
		tk_wai_flg(flgid, 0xffff, (TWF_ORW | TWF_CLR), &flgptn, TMO_FEVR);

		txchr = (flgptn >> 8) & 0xff;	// 送信データをtxchrに入れる(※R)
		rxchr = flgptn & 0xff;			// 受信データをrxchrに入れる

		tm_printf("flg_rx_task: dat=%04x, tx='%c', rx='%c'\n",
			(UINT) flgptn, disp_ch(txchr), disp_ch(rxchr));

		// 送信データと受信データのうち、データのある方をcmdchrに入れる(※W)
		if(txchr != 0){
			cmdchr = txchr;	// 送信データをcmdchrに入れる
			if(rxchr != 0)	// 送信完了と受信完了の割込みが同時に発生
							// この場合は受信データrxchrを無視するので警告
				tm_printf("flg_rx_task: rxchr(%c) ignored.\n", disp_ch(rxchr));
		} else if(rxchr != 0){
			cmdchr = rxchr;	// 受信データをcmdchrに入れる
		}

		if(cmdchr == 'A'){			// 次の曲に進んで演奏する(※X)
			mml_idx = (mml_idx + 1) % (sizeof(mml_main) / sizeof(B *));
			play_mml_idx();
		} else if(cmdchr == 'B'){	// 同じ曲をもう一度最初から演奏する
			play_mml_idx();
		}
	}
}

//---------------------------------------------------------------
// タスクのオブジェクトID番号
ID txtskid;				// 送信側タスク
ID rxtskid;				// 受信側タスク

// タスク、イベントフラグ、セマフォの生成情報
const T_CTSK ctxtsk = {0, (TA_HLNG | TA_RNG3), &btn_tx_task, 10, 1024, 0};
const T_CTSK crxtsk = {0, (TA_HLNG | TA_RNG3), &flg_rx_task, 10, 1024, 0};
const T_CFLG cflg = {0, (TA_TFIFO | TA_WSGL), 0};
const T_CSEM csem = {0, (TA_TFIFO | TA_FIRST ), 1, 1};

// 割込みハンドラ定義情報
const T_DINT dint_uarte = {TA_HLNG, uarte1_inthdr};

// UARTE1の割込み番号と割込み優先度レベル
#define INTNO_UARTE1	((UARTE1_BASE >> 12) & 0x3f)	// 割込み番号
#define INTPRI_UARTE1	6								// 割込み優先度レベル

// メインプログラム
EXPORT void usermain( void ){
	INT tx_pin, rx_pin;			// UARTEの送受信用エッジコネクタ端子の定義

	get_device_id();
	if(is_main_board()){		// メインボードとサブボードの判定(※S)
		tx_pin = 2; rx_pin = 3;	// メインボードの場合はP0TX_P1RX相当
	} else {
		tx_pin = 3; rx_pin = 2;	// メインボード以外の場合はP1TX_P0RX相当
	}

	btn_init();					// ボタンスイッチ入力のためのGPIOの初期設定

	// SPEAKERに接続されたGPIOピン(P0.00)を出力に設定
	out_w(GPIO(P0, PIN_CNF(SPEAKER_GPIO_PIN)), (1 << 0));

	// PWM(unit=0)のチャンネル0をGPIOピン(P0.00)経由でSPEAKERに出力
	out_w(PWM(speaker_pwm, PSEL_OUT(0)),
			((SPEAKER_GPIO_PORT << 5) | (SPEAKER_GPIO_PIN << 0)));

    led_init();						// LED制御用のGPIO設定

	// LEDを点灯させるためのGPIOの設定
	out_gpio_pin(0, 28, 0);		// GPIO P0.28-COL1に0を出力
	out_gpio_pin(0, 11, 0);		// GPIO P0.11-COL2に0を出力
	out_gpio_pin(0, 31, 0);		// GPIO P0.31-COL3に0を出力
	out_gpio_pin(1,  5, 0);		// GPIO P1.05-COL4に0を出力
	out_gpio_pin(0, 30, 0);		// GPIO P0.30-COL5に0を出力

	// LED用PWMの各チャンネル出力をGPIOピン経由でROW5..ROW2に出力
	out_w(PWM(led_pwm, PSEL_OUT(0)), ((0 << 5) | (19 << 0)));
										// チャンネル0→P0.19(ROW5に接続)
	out_w(PWM(led_pwm, PSEL_OUT(1)), ((0 << 5) | (24 << 0)));
										// チャンネル1→P0.24(ROW4に接続)
	out_w(PWM(led_pwm, PSEL_OUT(2)), ((0 << 5) | (15 << 0)));
										// チャンネル2→P0.15(ROW3に接続)
	out_w(PWM(led_pwm, PSEL_OUT(3)), ((0 << 5) | (22 << 0)));
										// チャンネル3→P0.22(ROW2に接続)

	init_pwm_duty();					// カウンタ比較値の配列データを設定

	init_uarte1(tx_pin, rx_pin);		// UARTE1の初期化と送受信用端子の指定
	uarte1_start_rx1();					// 最初の受信の準備

	flgid = tk_cre_flg(&cflg);			// 割込み通知用のイベントフラグ生成
	semid = tk_cre_sem(&csem);			// 送信バッファ排他制御用のセマフォ生成
	txtskid = tk_cre_tsk(&ctxtsk);		// 送信側タスクの生成
	rxtskid = tk_cre_tsk(&crxtsk);		// 受信側タスクの生成
	tk_sta_tsk(txtskid, 0);				// 送信側タスクの起動
	tk_sta_tsk(rxtskid, 0);				// 受信側タスクの起動

	tk_def_int(INTNO_UARTE1, &dint_uarte);	// UARTEの割込みハンドラ定義
	EnableInt(INTNO_UARTE1, INTPRI_UARTE1);	// 割込み許可

	tk_slp_tsk(TMO_FEVR);				// 永久待ち、以下は実行しない
}
